"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@pnpm/utils");
const loadJsonFile = require("load-json-file");
const path = require("path");
const writePkg = require("write-pkg");
function save(prefix, packageSpecs) {
    return __awaiter(this, void 0, void 0, function* () {
        // Read the latest version of package.json to avoid accidental overwriting
        let packageJson;
        const pkgJsonPath = path.join(prefix, 'package.json');
        try {
            packageJson = yield loadJsonFile(pkgJsonPath);
        }
        catch (err) {
            if (err['code'] !== 'ENOENT')
                throw err; // tslint:disable-line:no-string-literal
            packageJson = {};
        }
        packageSpecs.forEach((packageSpec) => {
            if (packageSpec.saveType) {
                const saveType = packageSpec.saveType;
                packageJson[packageSpec.saveType] = packageJson[packageSpec.saveType] || {};
                packageSpecs.forEach((dependency) => {
                    packageJson[saveType][dependency.name] = dependency.pref || findSpec(dependency.name, packageJson);
                    utils_1.dependenciesTypes.filter((deptype) => deptype !== packageSpec.saveType).forEach((deptype) => {
                        if (packageJson[deptype]) {
                            delete packageJson[deptype][dependency.name];
                        }
                    });
                });
            }
            else {
                packageSpecs.forEach((dependency) => {
                    const usedDepType = guessDependencyType(dependency.name, packageJson) || 'dependencies';
                    packageJson[usedDepType] = packageJson[usedDepType] || {};
                    packageJson[usedDepType][dependency.name] = dependency.pref;
                });
            }
        });
        yield writePkg(pkgJsonPath, packageJson);
        utils_1.packageJsonLogger.debug({
            prefix,
            updated: packageJson,
        });
        return packageJson;
    });
}
exports.default = save;
function findSpec(depName, pkg) {
    const foundDepType = guessDependencyType(depName, pkg);
    return foundDepType && pkg[foundDepType][depName];
}
function guessDependencyType(depName, pkg) {
    return utils_1.dependenciesTypes
        .find((deptype) => Boolean(pkg[deptype] && pkg[deptype][depName]));
}
//# sourceMappingURL=save.js.map